/*
 * cubic.c: simulate rubik's cube
 *         ("rubik's cube" is trade mark of Tsukuda Original)
 *
 * 1999.3, tsakurai@rimath.saitama-u.ac.jp
 * 1999.7, add mouse-action
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <GL/glut.h>
#include "vector.h"
#include "bmp.c"

#define DEG (3.14159265358979323846/180.0)

typedef struct _color4f {
    float r, g, b, a;
} Color4f;

typedef int IMatrix[3][3];

#define initIMatrix(M) \
{ IMatrix E={{1,0,0},{0,1,0},{0,0,1}}; memcpy(M, E, sizeof(IMatrix)); }

typedef struct _cubic {
    GLint   list;
    IMatrix M;
    Vector  center;
} Cubic;

/* Action */
typedef struct _action {
    unsigned char mask; /* 3 bits: selected slices */
      signed char axis; /* 0, +-1(x), +-2(y), +-3(z): rotation axis */
} Action;

#define STRDUP(S)  (strcpy((char *)malloc((unsigned)(strlen(S)+1)), (S)))

/*****************************************
 * glcubic global staff
 ----------------------------------------*/
/* main Object */
Cubic cubic_base[27], *cubic[27];

#define STACKSIZE 10000
Action act[STACKSIZE], *asp = act;

/* data file name */
char *dataFile = NULL;

/*******************************
 * global parameters 
 ------------------------------*/
/* flags */
int animation = 0;
int reading = 0;

/* stack tracing mode */
int trace_stack = 0;    /* 0: not in trace mode, 1: foreward, -1: backward */

/* eye = (r, theta, phi) in spherical coord */
double r = 30.0, theta = 70*DEG, phi = 30*DEG; 

/* rotation step */
double dangle = 90.0/8;

/* animation rot angle */
double zangle = 0;
#define dzangle (dangle/8)

/*****************************************
 * Polyhedron
 ----------------------------------------*/
typedef struct _Face {
    int    numVertex;
    Vector **varray;
    Vector normal;
    Color4f color;
} Face;

typedef struct _polyhedron {
    int numVertex;
    int numFace;
    Vector *vertex;
    Face   *face;
    Vector center;
} Polyhedron;

void PScale(Polyhedron *ph, double k)
{
    int i;
    for (i = 0; i < ph->numVertex; i++)
	VScale(&ph->vertex[i], &ph->vertex[i], k);
}

void PMove3f(Polyhedron *ph, double x, double y, double z)
{
    int i;
    VAdd3f(&ph->center, &ph->center, x, y, z);
    for (i = 0; i < ph->numVertex; i++)
	VAdd3f(&ph->vertex[i], &ph->vertex[i], x, y, z);
}

Vector getCenter(Face *f)
{
    Vector sum = {0, 0, 0};
    int i;
    for (i = 0; i < f->numVertex; i++)
	VAdd(&sum, &sum, f->varray[i]);
    VScale(&sum, &sum, 1.0/f->numVertex);
    return sum;
}

/*****************************************
 * setup Cubic
 ----------------------------------------*/
/* face color */ 
Color4f white  = { .7, .7, .8, 1.0};
Color4f green  = { .0, .5, .1, 1.0};
Color4f red    = { .6, .0, .0, 1.0};
Color4f orange = {1.0, .0, .0, 1.0};
Color4f blue   = { .0, .0, .5, 1.0};
Color4f yellow = { .8, .8, .0, 1.0};
Color4f gray   = { .1, .1, .1, 1.0};

static Polyhedron *newCube(void)
{
    Vector vx[8] = {{ 1,-1, 1}, { 1, 1, 1}, {-1, 1, 1}, {-1,-1, 1},
		    { 1,-1,-1}, { 1, 1,-1}, {-1, 1,-1}, {-1,-1,-1}};
    int  f[6][4] = {{0, 1, 2, 3}, {0, 4, 5, 1}, {1, 5, 6, 2},
		    {2, 6, 7, 3}, {3, 7, 4, 0}, {5, 4, 7, 6}};
    Vector  n[6] = {{ 0, 0, 1}, { 1, 0, 0}, { 0, 1, 0},
		    {-1, 0, 0}, { 0,-1, 0}, { 0, 0,-1}};
    Polyhedron *cube = (Polyhedron *)malloc(sizeof(Polyhedron)); {
	int  i, j;
	cube->numVertex = 8;
	cube->vertex = (Vector *)malloc(8*sizeof(Vector));
	for (i = 0; i < 8; i++)
	    cube->vertex[i] = vx[i];
	cube->numFace = 6;
	cube->face = (Face *)malloc(6*sizeof(Face));
	for (i = 0; i < 6; i++) {
	    cube->face[i].numVertex = 4;
	    cube->face[i].varray = (Vector **)malloc(4*sizeof(Vector *));
	    cube->face[i].normal = n[i];
	    cube->face[i].color  = gray;
	    for (j = 0; j < 4; j++)
		cube->face[i].varray[j] = cube->vertex + f[i][j];
	}
	VSet3f(&cube->center, .0, .0, .0);
    }
    return cube;
}

static void deleteCube(Polyhedron *cube)
{
    int  i;
    for (i = 0; i < 6; i++)
	free (cube->face[i].varray);
    free (cube->face);
    free (cube->vertex);
    free (cube);
}

static void initCube(Polyhedron **cube)
{
    int  i, j;
    for (i = 0; i < 27; i++) {
        PScale(cube[i], .45);
        PMove3f(cube[i], (i%3)-1.0, ((i%9)/3)-1.0, (i/9)-1.0);
	/* init face color */
	for (j = 0; j < 6; j++) {
            Face  *f  = &cube[i]->face[j];
            Vector fc = getCenter(f);
            if      (fc.z > 1.2)   f->color = white;
            else if (fc.z <-1.2)   f->color = blue;
            else if (fc.x > 1.2)   f->color = green;
            else if (fc.x <-1.2)   f->color = yellow;
            else if (fc.y > 1.2)   f->color = red;
            else if (fc.y <-1.2)   f->color = orange;
        }
    }
}

static void saveCubic(Polyhedron **cube)
{
    extern Cubic cubic_base[27], *cubic[27];
    GLfloat specular[] = {.9, .9, .9, 1.0},
            shininess  = .4*128.0;
    int  i, j, k;
    for (i = 0; i < 27; i++) {
	initIMatrix(cubic_base[i].M);
	cubic_base[i].center = cube[i]->center;
	cubic_base[i].list = glGenLists(1); {
	    glNewList(cubic_base[i].list, GL_COMPILE);
	    for (j = 0; j < 6; j++) {
		Face *f=&cube[i]->face[j];
		glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, &f->color.r);
		glMaterialfv(GL_FRONT, GL_SPECULAR,  specular);
		glMaterialf (GL_FRONT, GL_SHININESS, shininess);
		glBegin(GL_POLYGON); {
		    glNormal3dv(&f->normal.x);
		    for (k = 0; k < 4; k++) 
			glVertex3dv(&f->varray[k]->x);
		} glEnd();
	    }
	    glEndList();
	}
	cubic[i] = &cubic_base[i];
    }
}

void setupCubic(void)
{
    Polyhedron *cube[27];
    int i;
    for (i = 0; i < 27; i++) cube[i] = newCube();
    initCube(cube);
    saveCubic(cube);
    for (i = 0; i < 27; i++) deleteCube(cube[i]);
}

/*****************************************
 * Cubic action
 ----------------------------------------*/
static void glMultIMatrix(IMatrix M)
{
    double m[16] = {1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1};
    int  i, j;
    for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
	m[i + 4*j] = M[i][j];
    glMultMatrixd(m);
}

static void IMMult(IMatrix M, IMatrix A, IMatrix N)   /* M = AN */
{
    int i, j;
    IMatrix n; memcpy(n, N, sizeof(IMatrix));
    for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
	M[i][j] = A[i][0]*n[0][j] + A[i][1]*n[1][j] + A[i][2]*n[2][j];
}

static void reindex(void)
{
    int i;
    for (i = 0; i < 27; i++){
        Vector *cent = &cubic_base[i].center;
	int j, k, l;
        j = (int)((cent->x + 1.0) +.5); cent->x = j-1.0;
        k = (int)((cent->y + 1.0) +.5); cent->y = k-1.0;
        l = (int)((cent->z + 1.0) +.5); cent->z = l-1.0;
        cubic[j + 3*k + 9*l] = &cubic_base[i];
    }
}

#define sign(k) ((k > 0) ? 1 : -1)

#define PROJ(I, AXIS)   ((abs(AXIS) == 1)? ( (I) % 3 ): \
			 (abs(AXIS) == 2)? (((I)%9)/3): \
			 (abs(AXIS) == 3)? ( (I) / 9 ): 8)

void rotCubic(Action *a, int k)
{
    static Vector Ax_3[7] = {{ 0, 0,-1}, { 0,-1, 0}, {-1, 0, 0}, { 0, 0, 0},
			     { 1, 0, 0}, { 0, 1, 0}, { 0, 0, 1}};
    static Vector *Ax = Ax_3 + 3;
    static IMatrix R90_3[7] = {
    {{ 0, 1, 0}, {-1, 0, 0}, { 0, 0, 1}},   /* -z */
    {{ 0, 0,-1}, { 0, 1, 0}, { 1, 0, 0}},   /* -y */
    {{ 1, 0, 0}, { 0, 0, 1}, { 0,-1, 0}},   /* -x */
    {{ 1, 0, 0}, { 0, 1, 0}, { 0, 0, 1}},   /*  0 */
    {{ 1, 0, 0}, { 0, 0,-1}, { 0, 1, 0}},   /* +x */
    {{ 0, 0, 1}, { 0, 1, 0}, {-1, 0, 0}},   /* +y */
    {{ 0,-1, 0}, { 1, 0, 0}, { 0, 0, 1}},   /* +z */
    };
    static IMatrix *R90 = R90_3 + 3;
    double angle;
    int i;

    for (angle = 0; -90 <= angle && angle <= 90; angle += k*dangle) {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glPushMatrix();
	if (animation) glRotated(zangle+=dzangle, 0., 0.,-1.);
	for (i = 0; i < 27; i++) {
	    glPushMatrix();
	    if (a->mask & 1<<PROJ(i, a->axis))
		glRotated(angle, Ax[a->axis].x, Ax[a->axis].y, Ax[a->axis].z);
	    glMultIMatrix(cubic[i]->M);
	    glCallList(cubic[i]->list);
	    glPopMatrix();
	}
	glPopMatrix();
	glutSwapBuffers();
    }
    for (i = 0; i < 27; i++)
	if (a->mask & 1<<PROJ(i, a->axis)) {
	    IMMult(cubic[i]->M, R90[sign(k)*a->axis], cubic[i]->M);
	    MatVec(&cubic[i]->center, R90[sign(k)*a->axis], &cubic[i]->center)
	}
    reindex();
    glutPostRedisplay();
}

void scramble(int N)
{
    unsigned char mask[3] = {1, 2, 4};
    signed   char axis[6] = {1, 2, 3, -1, -2, -3};

    srand((unsigned) time(NULL));
    while (N--) {
        asp->mask = mask[(rand()/(RAND_MAX/9)%3)];
        asp->axis = axis[rand()/(RAND_MAX/9)%6];
	rotCubic(asp++, 3);
    }
}

void resetCubic(void)
{
    int  i;
    for (i = 0; i < 27; i++) {
	initIMatrix(cubic_base[i].M);
	VSet3f(&cubic_base[i].center, (i%3)-1, ((i%9)/3)-1, (i/9)-1);
	cubic[i] = &cubic_base[i];
    }
    trace_stack = reading = 0;
    asp = act;
}

/*****************************************
 * glut call-back functions
 ----------------------------------------*/
/* lighting */
GLfloat ambient [] = {.3, .3, .3, 1.0};
GLfloat light0_ambient [] = {.2, .2, .2, 1.0};
GLfloat light0_diffuse [] = {.6, .6, .6, 1.0};
GLfloat light0_specular[] = {.6, .6, .6, 1.0};
GLfloat light0_position[] = { 7.0,-2.0, 9.0, 0.0};

GLfloat light1_ambient [] = {.1, .1, .1, 1.0};
GLfloat light1_diffuse [] = {.3, .3, .3, 1.0};
GLfloat light1_specular[] = {.6, .6, .6, 1.0};
GLfloat light1_position[] = {-5.0, 5.0,-5.0, 0.0};

void myinit(void)
{
    /* set light properties */
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
        glLightfv(GL_LIGHT0, GL_AMBIENT,  light0_ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0_diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light0_specular);
    glEnable(GL_LIGHT1);
        glLightfv(GL_LIGHT1, GL_AMBIENT,  light1_ambient);
	glLightfv(GL_LIGHT1, GL_DIFFUSE,  light1_diffuse);
	glLightfv(GL_LIGHT1, GL_SPECULAR, light1_specular);

    /* Z-Buffer */
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);

    /* others */
    glShadeModel(GL_FLAT);
}

void draw(void)
{
    /* view port aspect ratio */
    double aspect =
	(double)glutGet(GLUT_WINDOW_WIDTH)/glutGet(GLUT_WINDOW_HEIGHT);

    /* set camera */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(14.0, aspect, 0.85*r, 1.2*r);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(r*sin(theta)*cos(phi), r*sin(theta)*sin(phi), r*cos(theta),
	      0.0, 0.0, 0.0,
	      -cos(theta)*cos(phi), -cos(theta)*sin(phi), sin(theta));

    /* put light */
    glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
    glLightfv(GL_LIGHT1, GL_POSITION, light1_position);

    /* draw */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glPushMatrix();
    if (animation) glRotated(zangle += dzangle, 0., 0.,-1.);
    /* cubic */ {
	int i;
	for (i = 0; i < 27; i++) {
	    glPushMatrix(); {
		glMultIMatrix(cubic[i]->M);
		glCallList(cubic[i]->list);
	    } glPopMatrix();
	}
    }
    glPopMatrix();
    glutSwapBuffers();
}

void idle(void)
{
    if (dataFile != NULL) {     /* trace file */
	void makeAction(int);
	static int c = EOF;     /* use EOF as reset flag */
	static FILE *fp;
	if (c == EOF && !(fp = fopen(dataFile, "r"))) {
	    fprintf(stderr,"cannot open '%s'", dataFile);
	    free(dataFile); dataFile = NULL; return;
	}
	if (reading && (c = getc(fp)) != EOF) {
	    if (c == '*') do {  /* entering scramble mode */ 
		if (c == '#') while (getc(fp) != '\n'); /* skip comment */
		makeAction(c);
	    } while ((c = getc(fp)) != '/');
	    if (c == '#') while (getc(fp) != '\n');     /* skip comment */
	    makeAction(c);
	} else {
	    fclose(fp); free(dataFile); dataFile = NULL;
	    reading = 0; c = EOF;
	}
    } else {                    /* trace stack */
	if (trace_stack > 0)
	    if (asp->axis != 0) rotCubic(asp++, 1); else trace_stack = 0;
	else if (trace_stack < 0)
	    if (asp > act)      rotCubic(--asp,-1); else trace_stack = 0;
    }
    if (animation) glutPostRedisplay();
}

Action getAction(Vector *pt0, Vector *mv)
{
    static const unsigned char mask[3] = {1, 2, 4};
    Action a;

    if (fabs(fabs(pt0->x) - 1.45) < .03) {
	if (fabs(mv->y) > fabs(mv->z)) {
	    a.axis = (pt0->x*mv->y > 0) ? 3 : -3;
	    a.mask = mask[(int)(pt0->z+1.5)];
	} else {
	    a.axis = (pt0->x*mv->z < 0) ? 2 : -2;
	    a.mask = mask[(int)(pt0->y+1.5)];
	}
    } else if (fabs(fabs(pt0->y) - 1.45) < .03) {
	if (fabs(mv->z) > fabs(mv->x)) {
	    a.axis = (pt0->y*mv->z > 0) ? 1 : -1;
	    a.mask = mask[(int)(pt0->x+1.5)];
	} else {
	    a.axis = (pt0->y*mv->x < 0) ? 3 : -3;
	    a.mask = mask[(int)(pt0->z+1.5)];
	}
    } else if (fabs(fabs(pt0->z) - 1.45) < .03) {
	if (fabs(mv->x) > fabs(mv->y)) {
	    a.axis = (pt0->z*mv->x > 0) ? 2: -2;
	    a.mask = mask[(int)(pt0->y+1.5)];
	} else {
	    a.axis = (pt0->z*mv->y < 0) ? 1 : -1;
	    a.mask = mask[(int)(pt0->x+1.5)];
	}
    }
    return a;
}

#define GLUT_MOVE  2    /* GLUT_DOWN = 0, GLUT_UP = 1 */

void mouse(int button, int state, int x, int y)
{
    static GLdouble mvmatrix[16], projmatrix[16];
    static GLint viewport[4];
    static Vector pt, pt0, mv;
    static int selecting = 0;
    GLfloat z; /* depth value */

    if (button != GLUT_LEFT_BUTTON) { return; }
    switch (state) {
    case GLUT_DOWN:
        glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix);
        glGetDoublev(GL_PROJECTION_MATRIX, projmatrix);
        glGetIntegerv(GL_VIEWPORT, viewport);
        y = viewport[3] - 1 - y;  /* mouse pos to raster pos */
        glReadPixels(x, y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &z);
	if (z == 1.0) {
	    asp->mask = 7;
	    if (abs(x-viewport[2]/2) > abs(y-viewport[3]/2)) 
		asp->axis = (x-viewport[2]/2 > 0) ? 3 : -3;
	    else
		asp->axis = (y-viewport[3]/2 < 0) ? 2 : -2;
	    rotCubic(asp++, 1);
	    return;
	}
	gluUnProject(x, y, (double)z, mvmatrix, projmatrix, viewport,
		     &pt.x, &pt.y, &pt.z);
        pt0 = pt;
        selecting = 1;
        return;
    case GLUT_MOVE:
        if (selecting == 0) return;
        y = viewport[3] - 1 - y;  /* mouse pos to raster pos */
        glReadPixels(x, y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &z);
	gluUnProject(x, y, (double)z, mvmatrix, projmatrix, viewport,
                         &pt.x, &pt.y, &pt.z);
	VSub(&mv, &pt, &pt0);
	if (VLength(&mv) < .1) return;
	*asp = getAction(&pt0, &mv);
	rotCubic(asp++, 1);
        selecting = 0;
        return;
    case GLUT_UP:
	if (selecting && asp > act) rotCubic(--asp, -1);
        selecting = 0;
        return;
    default:
        return;
    }
}

void motion(int x, int y) { mouse(GLUT_LEFT_BUTTON, GLUT_MOVE, x, y); }

void makeAction(int c)
{
    void writeLog(char *);
    static Action a;

    switch (c) {
    case 'q':  writeLog("cubic.log"); exit(0);
    case '0':  a.mask=0; break;
    case '1':  a.mask=1; break;
    case '2':  a.mask=2; break;
    case '3':  a.mask=4; break;
    case '4':  a.mask=7; break;
    case ' ':  a.mask=7; break;
    case 'j':  a.axis= 1; *asp=a; rotCubic(asp++,1); break;
    case 'k':  a.axis= 2; *asp=a; rotCubic(asp++,1); break;
    case 'l':  a.axis= 3; *asp=a; rotCubic(asp++,1); break;
    case 'u':  a.axis=-1; *asp=a; rotCubic(asp++,1); break;
    case 'i':  a.axis=-2; *asp=a; rotCubic(asp++,1); break;
    case 'o':  a.axis=-3; *asp=a; rotCubic(asp++,1); break;
    case 'r':  resetCubic(); break;
    case 's':  scramble(30); break;
    case 'b':  if (asp > act) rotCubic(--asp,-1); break;
    case 'f':  if (asp->axis != 0) rotCubic(asp++,1); break;
    case 'B':  trace_stack =-1; break;
    case 'F':  trace_stack = 1; break;
    case 'D':  r *=.9; break;
    case 'd':  r /=.9; break;
    case 'a':  animation = !animation; zangle = 0; break;
    case '*':  dangle *= 2; break;
    case ':':  dangle *= 2; break;
    case '/':  dangle /= 2; break;
    case 'S':  draw(); glReadBuffer(GL_FRONT); SaveImageAs("cubic.bmp"); break;
    case 1002: dangle = 90.0/  4; break;
    case 1003: dangle = 90.0/  8; break;
    case 1004: dangle = 90.0/ 16; break;
    case 1005: dangle = 90.0/ 32; break;
    case 1006: dangle = 90.0/ 64; break;
    case 1007: dangle = 90.0/128; break;
    case 2000: resetCubic(); dataFile = STRDUP("00.dat"); reading = 1; break;
    case 2001: resetCubic(); dataFile = STRDUP("01.dat"); reading = 1; break;
    case 2002: resetCubic(); dataFile = STRDUP("02.dat"); reading = 1; break;
    default:   return;
    }
    glutPostRedisplay();
}

void key(unsigned char c, int x, int y) { makeAction((int) c); }

void skey(int k, int x, int y)
{
    switch (k) {
    case GLUT_KEY_UP:	 theta -= 2*DEG; break;
    case GLUT_KEY_DOWN:	 theta += 2*DEG; break;
    case GLUT_KEY_LEFT:	 phi   -= 2*DEG; break;
    case GLUT_KEY_RIGHT: phi   += 2*DEG; break;
    default: return;
    }
    glutPostRedisplay();
}

void makeMenu(void)
{
    GLint demo, rot_step;
    rot_step = glutCreateMenu(makeAction); {
        glutAddMenuEntry("  4",           1002);
        glutAddMenuEntry("  8 (default)", 1003);
        glutAddMenuEntry(" 16",           1004);
        glutAddMenuEntry(" 32",           1005);
        glutAddMenuEntry(" 64",           1006);
        glutAddMenuEntry("128",           1007);
    }
    demo = glutCreateMenu(makeAction); {
        glutAddMenuEntry(" 00.dat",       2000);
        glutAddMenuEntry(" 01.dat",       2001);
        glutAddMenuEntry(" 02.dat",       2002);
    }
    glutCreateMenu(makeAction); {
        glutAddMenuEntry("Undo",             'b');
        glutAddMenuEntry("Redo",             'f');
        glutAddMenuEntry("Scramble",         's');
        glutAddMenuEntry("Reset Cube",       'r');
        glutAddMenuEntry("Trace Back",       'B');
        glutAddMenuEntry("Toggle Animation", 'a');
        glutAddMenuEntry("Close up",         'D');
        glutAddMenuEntry("Move away",        'd');
        glutAddSubMenu("Rotation Step", rot_step);
        glutAddSubMenu("Demonstration",     demo);
        glutAddMenuEntry("Save Image (cubic.bmp)", 'S');
        glutAddMenuEntry("Quit",             'q');
    }
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

void reshape(int width, int height) { glViewport(0, 0, width, height); }
void visible(int vis) { glutIdleFunc((vis == GLUT_VISIBLE) ? idle : NULL); }

/*****************************************
 * main
 ----------------------------------------*/
void usage(void)
{
    fprintf(stderr, 
	    "\n"
	    "cubic [-a] [-s n] [file]\n"
	    "\n"
	    "-a    animetion on\n"
	    "-s n  rot_step = 2^n\n"
	    "file  read action from the file\n"
	    "\n"
	    "Keys:\n"
	    "   1, 2, 3 : select 'slice'\n"
	    "   4, space: all\n"
	    "   j, k, l : rotate cubes\n"
            "   u, i, o : j,u = x; k,i = y; l,o = z\n"
	    "   s       : scramble\n"
	    "   b       : back 1 step (Undo)\n"
	    "\n");
}

int main(int argc, char *argv[])
{
    int idx;
    for (idx = 1; idx < argc; idx++)
	if (*argv[idx] == '-') {
	    if (strcmp(argv[idx], "-a") == 0) {	animation = 1; continue; }
	    if (strcmp(argv[idx], "-s") == 0) {
		int exp; /* default: 3 */
		if (sscanf(argv[++idx], "%d", &exp) != 1) {
		    fprintf(stderr, "missing step number"); usage(); exit(1);
		}
		dangle = 90.0/(1<<exp); continue;
	    }
	    fprintf(stderr, "illegal option '%s'\n", argv[idx]);
	    usage(); exit(1);
	} else {	/* Should be an input file */
	    dataFile = STRDUP(argv[idx]); reading = 1;
	}

    /* initialize */
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
    glutInitWindowPosition(0, 0);
    glutInitWindowSize(500, 400);
    glutCreateWindow(argv[0]);
    myinit();

    /* setup Cubes */
    setupCubic();
    
    /* regist call-back functions */
    glutDisplayFunc(draw);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(key);
    glutSpecialFunc(skey);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
    glutVisibilityFunc(visible);
    makeMenu();
    
    glutMainLoop();
    return 0;
}

/*****************************************
 * write log
 ----------------------------------------*/
void writeLog(char *fname)
{
    char *smask[8] = {"0", "1", "2", "12", "3", "13", "23", "\n4"};
    char *c_3[7] = {"o", "i", "u", "", "j", "k", "l"}, **c = c_3 + 3;
    Action *ap;
    FILE *fp;
    if (!(fp = fopen(fname, "w"))) return;
    for (ap = act; ap < asp; ap++) {
	if (ap->mask == 0) ;
	else if (ap->mask == (ap-1)->mask)
	    fprintf(fp, "%s", c[ap->axis]);
	else
	    fprintf(fp, " %s%s", smask[ap->mask], c[ap->axis]);
    }
    fclose(fp);
}
